#!/usr/bin/env python

"""
PeaKO ranks motifs based on peak overlap between differential
MEME-ChIP (DMC) and knockout implemented normalization (KOIN)
pipeline peak sets.

Note: In our manuscript, we apply the following mappings:
     DMC pipeline --> Pipeline A
     KOIN pipeline --> Pipeline B
"""

from __future__ import absolute_import, division, print_function
import argparse
import linecache
from math import log
from collections import OrderedDict
import csv
import os
from itertools import chain
import logging

import yaml
from bs4 import BeautifulSoup
import pandas as pd
import pybedtools
import pickle
from six.moves import range


class CentriMoOutput:

    def __init__(self, filename):
        self.filename = filename

    def data(self):
        return store_data_from_html(self.filename)

    def num_db_motifs(self):
        """
        Number of motifs in provided databases
        (usually JASPAR, MEME, and/or DREME).
        """
        return sum([db['count'] for db in self.data()['motif_dbs']])

    def num_motifs(self):
        """
        Number of motifs that CentriMo found to be sufficiently
        enriched in peak sets.
        """
        return len(self.data()['motifs'])


def store_data_from_html(filename_html):
    """
    Return dictionary of HTML file between script tags.

    Use Beautiful Soup module from Python to extract text between script tags
    in HTML file. Then, split on '=' and remove semicolons.

    We are extracting the set of all motifs reported by CentriMo.


    Keyword arguments:
    filename_html -- full path of HTML file
    """
    with open(filename_html, 'r') as f:
        html = f.read()
        soup = BeautifulSoup(html, 'html.parser')
        script = soup.find('script')
        # adapted from: http://stackoverflow.com/questions/13323976/
        json_text = script.text.split('=', 1)[1].strip().rstrip(';')
        data = yaml.safe_load(json_text)  # loads into string objects
        return data


def read_centrimo_df(filename_txt):
    """
    Read CentriMo TXT file and return dataframe.
    """
    line = linecache.getline(filename_txt, 2)  # line 1: warning
    header = line.split()[1:]  # split on whitespace and ignore leading '#'
    df = pd.read_csv(filename_txt, header=None, names=header,
                     skip_blank_lines=True, comment='#', delim_whitespace=True)
    return df


def read_and_sort_centrimo_df(filename_txt):
    """
    Read CentriMo TXT file and return dataframe sorted by (log-adjusted)
    binomial p-values (most significant motifs at top).


    Keyword arguments:
    filename_txt -- full path of CentriMo TXT file
    """
    df = read_centrimo_df(filename_txt)
    sorted_df = (df
                 .sort_values(by='log_adj_p-value', ascending=True)
                 .reset_index(drop=True))
    sorted_df.index += 1
    return sorted_df


def read_and_sort_fisher_exact_centrimo_df(filename_txt):  # used in counts
    """
    Read differential CentriMo TXT file and return dataframe sorted by
    (log-adjusted) Fisher's Exact test p-values (most significant motifs
    at top).

    Key word arguments:
    filename_txt -- full path of differential CentriMo TXT file
    """
    df = read_centrimo_df(filename_txt)
    sorted_df = (df
                 .sort_values(by='log_fisher_adj_pvalue', ascending=True)
                 .reset_index(drop=True))
    sorted_df.index += 1
    return sorted_df


def filter_centrimo_df(df, CentriMoOutput_):
    """
    Filter out motifs with E-values greater than 0.1.


    Keyword arguments:
    df -- sorted dataframe of CentriMo TXT values
    CentriMoOutput_ -- CentriMoOutput class object
    """
    log_adj_p_val_filter = log(0.1/CentriMoOutput_.num_db_motifs())
    filtered_motif_list = (df
                           .loc[df['log_adj_p-value'] <=
                                log_adj_p_val_filter]['id'])
    return filtered_motif_list


def write_dict_to_bed_file(dictionary, outdir, file_list, filetype):
    """
    Write contents of peak set dictionary to BED file.


    Keyword arguments:
    dictionary -- keys are motifs, values are genomic region intervals
    file_list -- list to append newly created BED files
    filetype -- one of 'intersection', 'koin', 'neg', 'pos'
    """
    for motif, regions in dictionary.items():

        if filetype == 'intersection':
            bed_file = os.path.join(outdir, 'intcount-{}.bed'.format(motif))
            file_list.append(bed_file)
            with open(bed_file, 'w') as csv_file:
                writer = csv.writer(csv_file, delimiter='\t',
                                    lineterminator='\n')
                for reg_index in range(0, len(regions)):
                    region = regions[reg_index]
                    writer.writerow([region[0], region[1], region[2]])
        else:
            if filetype == 'koin':
                bed_file = os.path.join(outdir, 'koin-{}.bed'.format(motif))
            elif filetype == 'neg':
                bed_file = os.path.join(outdir, '{}-neg.bed'.format(motif))
            elif filetype == 'pos':
                bed_file = os.path.join(outdir, '{}.bed'.format(motif))

            file_list.append(bed_file)
            with open(bed_file, 'w') as csv_file:
                writer = csv.writer(csv_file, delimiter='\t',
                                    lineterminator='\n')
                for reg_index in range(0, len(regions)):
                    chrm, region = regions[reg_index].split(':')
                    start, end = region.split('-')
                    writer.writerow([chrm, int(start), int(end)])
        logging.debug('Wrote {} file.'
                      .format(os.path.basename(bed_file)))


def pickle_dictionary(dictionary, pickle_file):
    with open(pickle_file, 'wb') as handle:
        pickle.dump(dictionary, handle)


def load_pickled_dict(pickle_file):
    with open(pickle_file, 'rb') as handle:
        return pickle.load(handle)


def remove_prefix(text, prefix):
    # adapted from https://stackoverflow.com/questions/16891340
    if text.startswith(prefix):
        return text[len(prefix):]


def main():
    logging.basicConfig(format='%(asctime)s >peaKO: %(message)s',
                        datefmt='%m/%d/%Y %I:%M:%S %p', level=logging.INFO)

    parser = argparse.ArgumentParser(
        prog='peako',
        description="""
        PeaKO ranks motifs from ChIP-seq experiments performed with knockout
        controls. PeaKO computes its ranking metric by comparing CentriMo
        results across two different ways of differentially accounting for
        the knockout dataset (Pipeline A vs. Pipeline B; see manuscript
        for details).

        """,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""

Example CentriMo HTML and TXT files are available at: \
https://doi.org/10.5281/zenodo.3338330.

Usage example:
$ ./peako.py ~/workdir/centrimo/centrimo-a/ \

             centrimo.html                  \

             centrimo.txt                   \

             ~/workdir/centrimo/centrimo-b/ \

             centrimo.html                  \

             centrimo.txt                   \

             ~/workdir/peako/

If you found peaKO useful, please cite:

Denisko D, Viner C, Hoffman MM. Motif elucidation in ChIP-seq datasets with a
knockout control. BioRxiv <ID> [Preprint]. 2019. \
Available from: https://doi.org/<ID>
        """)
    # req_grp = parser.add_argument_group(title='required arguments')
    parser.add_argument('indir_a', metavar='pipeline-a-dir',
                        help='differential CentriMo directory ' +
                        '(Pipeline A)')
    parser.add_argument('html_a', metavar='pipeline-a-html',
                        help='differential CentriMo HTML filename ' +
                        '(Pipeline A)')
    parser.add_argument('txt_a', metavar='pipeline-a-txt',
                        help='differential CentriMo TXT filename ' +
                        '(Pipeline A)')
    parser.add_argument('indir_b', metavar='pipeline-b-dir',
                        help='non-differential CentriMo directory ' +
                        '(Pipeline B)')  # KOIN
    parser.add_argument('html_b', metavar='pipeline-b-html',
                        help='non-differential CentriMo HTML filename ' +
                        '(Pipeline B)')  # KOIN
    parser.add_argument('txt_b', metavar='pipeline-b-txt',
                        help='non-differential CentriMo TXT filename ' +
                        '(Pipeline B)')  # KOIN
    parser.add_argument('outdir', help='output directory')
    parser.add_argument('-V', '--version', action='version',
                        version='%(prog)s 1.0')
    parser.add_argument('-v', '--verbose', action="store_true",
                        help='print detailed progress')
    parser.add_argument('-m', dest='motif',
                        help='transcription factor motif common name ' +
                        '(e.g. SRF)')
    parser.add_argument('-j', dest='jaspar_id',
                        help='transcription factor motif JASPAR ' +
                        'identifier (e.g. MA0083.3)')
    parser.add_argument('--pickled', action="store_true",
                        help='use pickled dictionaries from previous run(s) ' +
                        'located in the output directory (outdir)')
    parser.add_argument('--extra', action="store_true",
                        help='output all intermediate files for plotting')
    parser.add_argument('--filter', action="store_true",
                        help='filter out half of peaKO motifs, according to ' +
                        'the number of intersecting regions')

    args = parser.parse_args()
    logging.info('Command line arguments: {}'.format(args))

    if args.verbose:
        logging.getLogger().setLevel(logging.DEBUG)

    # dmc_html = os.path.join(args.d, args.d1)
    dmc_html = os.path.join(args.indir_a, args.html_a)
    # dmc_txt = os.path.join(args.d, args.d2)
    dmc_txt = os.path.join(args.indir_a, args.txt_a)
    # koin_html = os.path.join(args.k, args.k1)
    koin_html = os.path.join(args.indir_b, args.html_b)
    # koin_txt = os.path.join(args.k, args.k2)
    koin_txt = os.path.join(args.indir_b, args.txt_b)
    outdir = args.outdir
    transcription_factor = args.motif
    jaspar_id = args.jaspar_id
    pickled = args.pickled
    extra = args.extra

    # Create CentriMoOutput type objects to store HTML.
    dmc_html_obj = CentriMoOutput(dmc_html)
    koin_html_obj = CentriMoOutput(koin_html)
    logging.info('Read HTML files.')

    # Create dataframes from TXT files and sort according to log-adjusted
    # (binomial) p-values.
    sorted_dmc_df = read_and_sort_centrimo_df(dmc_txt)
    sorted_koin_df = read_and_sort_centrimo_df(koin_txt)

    # Filter out motifs with E value > 0.1 and
    # append those remaining to a list.
    filtered_dmc_motif_list = filter_centrimo_df(sorted_dmc_df,
                                                 dmc_html_obj).tolist()
    filtered_koin_motif_list = filter_centrimo_df(sorted_koin_df,
                                                  koin_html_obj).tolist()
    logging.info('Created and sorted dataframes. Filtered out motifs'
                 ' with E > 0.1. Pipeline A list contains {} motifs.'
                 ' Pipeline B list contains {} motifs.'.format(
                     len(filtered_dmc_motif_list),
                     len(filtered_koin_motif_list)))

    # Create dictionaries:
    # keys: motifs passing E-value filtering
    # values: chromosome regions where these motifs bind
    #         (as determined by CentriMo)
    #         (format: lists containing 'chri:xxxxxxx-xxxxx5xx' strings)

    dct = OrderedDict()
    neg_dct = OrderedDict()
    koin_dct = OrderedDict()

    # Calculate total number of motifs and make lists of motif IDs from
    # original HTML files.
    number_of_motifs = dmc_html_obj.num_motifs()
    dmc_pos_html_motifs = [d['id'] for d in dmc_html_obj.data()['motifs']]
    number_of_koin_motifs = koin_html_obj.num_motifs()
    koin_html_motifs = [d['id'] for d in koin_html_obj.data()['motifs']]
    logging.info('Calculated total number of motifs and created ID lists.')

    # Make dictionaries to store motifs or load in pickled dicts
    dct_pickle_file = os.path.join(outdir, 'dct.pickle')
    neg_dct_pickle_file = os.path.join(outdir, 'neg_dct.pickle')
    koin_dct_pickle_file = os.path.join(outdir, 'koin_dct.pickle')

    if not pickled:
        # Create dictionaries.
        logging.warning('No pickled dictionaries, so creating some.')

        # Make lists of all sequences from HTMLs.
        pos_sequence_list = dmc_html_obj.data()['sequences']
        neg_sequence_list = dmc_html_obj.data()['neg_sequences']
        koin_sequence_list = koin_html_obj.data()['sequences']
        logging.info('Created sequence lists.')

        for filtered_dmc_motif in filtered_dmc_motif_list:

            for index, motif in enumerate(dmc_pos_html_motifs):

                if motif == filtered_dmc_motif:
                    logging.debug('Creating dict entry for {}...'.format(
                        motif))
                    # array of integers (indices) correspond
                    # to regions in dmc_html_obj.data()['sequences']
                    motif_dct = dmc_html_obj.data()['motifs'][index]

                    for seqnum in motif_dct['seqs']:
                        if "id{}".format(filtered_dmc_motif) in dct:
                            # adapted from http://stackoverflow.com/
                            # questions/3199171
                            dct["id{}"
                                .format(filtered_dmc_motif)].append(
                                    pos_sequence_list[seqnum])
                        else:
                            dct["id{}"
                                .format(filtered_dmc_motif)] = [
                                    pos_sequence_list[seqnum]]

                    for neg_seqnum in motif_dct['neg_seqs']:
                        if "id{}".format(filtered_dmc_motif) in neg_dct:
                            neg_dct["id{}".format(filtered_dmc_motif)].append(
                                neg_sequence_list[neg_seqnum])
                        else:
                            neg_dct["id{}".format(filtered_dmc_motif)] = [
                                neg_sequence_list[neg_seqnum]]
                        break
        logging.info('Created Pipeline A positive and negative sequence ' +
                     'dictionaries...')

        for filtered_koin_motif in filtered_koin_motif_list:

            for index, motif in enumerate(koin_html_motifs):

                if motif == filtered_koin_motif:
                    logging.debug('Creating Pipeline B dict entry ' +
                                  'for {}...'.format(motif))
                    motif_dct = koin_html_obj.data()['motifs'][index]

                    for seqnum in motif_dct['seqs']:
                        if "id{}".format(filtered_koin_motif) in koin_dct:
                            koin_dct["id{}".format(
                                filtered_koin_motif)].append(
                                    koin_sequence_list[seqnum])
                        else:
                            koin_dct["id{}".format(filtered_koin_motif)] = [
                                koin_sequence_list[seqnum]]
                    break
        logging.info('... and Pipeline B sequence dictionaries...')
        logging.info('Created dictionaries to store motifs and '
                     'their associated chromosome regions.')

        # Create 3-column BED files for motifs in each dictionary.
        pickle_dictionary(dct, dct_pickle_file)
        pickle_dictionary(neg_dct, neg_dct_pickle_file)
        pickle_dictionary(koin_dct, koin_dct_pickle_file)

        logging.info('Pickled dictionaries. ')

    elif pickled:
        logging.info('Loading pickled dictionaries...')
        # to reload pickled objects
        dct = load_pickled_dict(dct_pickle_file)
        neg_dct = load_pickled_dict(neg_dct_pickle_file)
        koin_dct = load_pickled_dict(koin_dct_pickle_file)
        logging.info('Loaded!')
    # else:
        # logging.warning('Neither pickled nor not pickled...')

    dmc_file_list = list()
    negMCfile_list = list()
    koinfile_list = list()

    write_dict_to_bed_file(dct, outdir, dmc_file_list, 'pos')
    write_dict_to_bed_file(neg_dct, outdir, negMCfile_list, 'neg')
    write_dict_to_bed_file(koin_dct, outdir, koinfile_list, 'koin')

    logging.info('Wrote dictionaries to BED files. ')

    # Calculate ratio score and print results to file.

    intersection_regs_dct = OrderedDict()
    peako_file_unsorted = os.path.join(outdir, 'peaKO-rankings-unsorted.txt')

    with open(peako_file_unsorted, 'w') as csv_file:
        writer = csv.writer(csv_file, delimiter='\t', lineterminator='\n')
        # number_of_intersection_regs, number_of_DMC_positive_regs,
        # number_of_K_regs
        writer.writerow(["motif_ID", "intersection_regions",
                         "pipeline_a_wt_regions", "pipeline_b_regions",
                         "ratio", "alt_ID"])

        for dmc_file in dmc_file_list:
            dmc_pos_bedtool = pybedtools.BedTool(dmc_file)
#            motif_id = os.path.basename(dmc_file)[:-4]  # assume *.bed # XXX
            dmc_base, dmc_ext = os.path.splitext(dmc_file)
            assert dmc_ext == ".bed"
            motif_id = os.path.basename(dmc_base)
            dmc_neg_file = os.path.join(outdir, '{}-neg.bed'.format(motif_id))

            for koin_file in koinfile_list:
                koin_base, koin_ext = os.path.splitext(koin_file)
                assert koin_ext == ".bed"
                if remove_prefix(os.path.basename(koin_base),
                                 'koin-') == motif_id:
                    # if os.path.basename(koin_file)[5:-4] == motif_id:
                    koin_bedtool = pybedtools.BedTool(koin_file)

                    try:
                        dmc_neg_bedtool = pybedtools.BedTool(dmc_neg_file)

                        # 'A=True' removes full feature in dmc_pos_bedtool
                        # if overlapping
                        dmc_difference = dmc_pos_bedtool.subtract(
                            dmc_neg_bedtool, A=True)

                        # 'wa=True' reports full feature of dmc_difference
                        # if overlapping
                        koin_intersect_dmc = (dmc_difference
                                              .intersect(koin_bedtool,
                                                         wa=True))
                    # workaround for no negative file
                    except (FileNotFoundError, ValueError):
                        logging.warning('The negative sequences bed file,'
                                        ' {}, does not exist.'
                                        .format(os.path.basename(
                                            dmc_neg_file)))
                        koin_intersect_dmc = (dmc_pos_bedtool
                                              .intersect(koin_bedtool,
                                                         wa=True))

                    koin_intersect_dmc_merged = (koin_intersect_dmc
                                                 .sort()
                                                 .merge())
                    number_of_intersection_regs = (koin_intersect_dmc_merged
                                                   .count())
                    logging.debug('Motif {} has {} intersection ' +
                                  'regions'.format(
                                      motif_id, number_of_intersection_regs))

                    # Create dictionary entry to print intersecting regions
                    # for each motif.
                    for interval in koin_intersect_dmc_merged:
                        if "{}".format(motif_id) in intersection_regs_dct:
                            (intersection_regs_dct["{}".format(motif_id)]
                             .append(interval))
                        else:
                            intersection_regs_dct["{}"
                                                  .format(
                                                      motif_id)] = [interval]

                    # Find alternate ID using original HTML file.
                    for motif in dmc_html_obj.data()['motifs']:
                        if motif_id == ('id' + motif['id']):
                            try:
                                motif_id_alt = motif['alt']
                            except (FileNotFoundError, ValueError):
                                motif_id_alt = motif['id']

                            ratio = float(
                                number_of_intersection_regs)/len(dct[motif_id])
                            writer.writerow([motif_id,
                                             number_of_intersection_regs,
                                             len(dct[motif_id]),
                                             len(koin_dct[motif_id]),
                                             ratio, motif_id_alt])

    if not extra:
        for dmc_file in dmc_file_list:
            os.remove(dmc_file)
            dmc_basename, dmc_ext = os.path.splitext(
                os.path.basename(dmc_file))
            dmc_neg_file = os.path.join(os.path.dirname(dmc_file),
                                        dmc_basename + '-neg' + dmc_ext)
            if os.path.isfile(dmc_neg_file):
                os.remove(dmc_neg_file)
        for koin_file in koinfile_list:
            os.remove(koin_file)
    logging.info('Wrote first (unsorted and unfiltered) ratio file.')

    # Print out intersection regions.
    if extra:
        intersection_regs_list = list()
        write_dict_to_bed_file(intersection_regs_dct, outdir,
                               intersection_regs_list, 'intersection')
        logging.info('Wrote BED files of intersection regions.')

        intersection_counts_list = (set(list(chain
                                             .from_iterable(
                                                 list(intersection_regs_dct.values())))))
        intersection_counts_file = os.path.join(outdir,
                                                'all-intcount-regions.bed')

        with open(intersection_counts_file, 'w') as csv_file:
            writer = csv.writer(csv_file, delimiter='\t', lineterminator='\n')
            for region in intersection_counts_list:
                writer.writerow([region[0], region[1], region[2]])
        logging.info('Wrote all intersection regions to file.')

    # Filter out the bottom 50% of motifs when ranked by the number
    # of intersecting regions.
    ratio_filepath = peako_file_unsorted
    rankings_df = pd.read_csv(ratio_filepath, delim_whitespace=True)
    os.remove(peako_file_unsorted)
    rankings_ratio_sorted_df = (rankings_df
                                .sort_values(by=['ratio',
                                                 'intersection_regions',
                                                 'pipeline_a_wt_regions',
                                                 'pipeline_b_regions'],
                                             ascending=False)
                                .reset_index(drop=True))
    rankings_ratio_sorted_df.index += 1

    peako_file = os.path.join(outdir, 'peaKO-rankings.txt')
    rankings_ratio_sorted_df.to_csv(peako_file,
                                    index=True, sep='\t')
    logging.info('Wrote peaKO file (sorted).')

    filtering_threshold = int(len(rankings_df)/2)  # 50% of motifs
    rankings_intcount_filtered_df = (rankings_df
                                     .sort_values(by=['intersection_regions',
                                                      'ratio'],
                                                  ascending=False)[
                                                    :filtering_threshold]
                                     .sort_values(by=['ratio',
                                                      'intersection_regions',
                                                      'pipeline_a_wt_regions',
                                                      'pipeline_b_regions'],
                                                  ascending=False)
                                     .reset_index(drop=True))
    rankings_intcount_filtered_df.index += 1

    if args.filter:
        peako_file_filtered = os.path.join(outdir,
                                           'peaKO-rankings-filtered.txt')
        rankings_intcount_filtered_df.to_csv(peako_file_filtered, index=True,
                                             sep='\t')
        logging.info('Wrote peaKO file filtered (top 50% by number of ' +
                     'intersecting regions).')

    # find and print motif's original vs. new ranking
    if jaspar_id or transcription_factor:
        if jaspar_id:
            logging.info('Using JASPAR ID ({}) to '.format(jaspar_id) +
                         'find ranks.')

            label = jaspar_id
            my_regex = "id" + jaspar_id
            column = "motif_ID"

        elif transcription_factor:
            logging.info('Using motif common name ({}) to find ranks.'.format(
                    transcription_factor))

            label = transcription_factor
            my_regex = "(?i)^" + transcription_factor + "$"  # case insensitive
            column = "alt_ID"

        # record motif's ranking as index (peaKO without filtering)
        peako_rank = (rankings_ratio_sorted_df[rankings_ratio_sorted_df[column]
                                               .str.contains(my_regex)]
                      .index[0])

        logging.info('PeaKO ranks {} in position {}.'
                     .format(label, peako_rank))

        try:
            peako_with_filtering_rank = (rankings_intcount_filtered_df[
                rankings_intcount_filtered_df[column]
                .str.contains(my_regex)]
                                         .index[0])
        except IndexError:
            peako_with_filtering_rank = 'NA (motif has been filtered out)'

        if args.filter:
            logging.info('PeaKO after 50% filtering by number of ' +
                         'intersections ranks {} in position {}.'
                         .format(label, peako_with_filtering_rank))
    else:
        logging.warning('User did not provide any motif IDs.')

    if extra:
        # print all rankings to file to use in plotting
        # the user can manually "cat" all output files afterwards
        logging.info('Generating CSV file to be used for plotting...')

        if jaspar_id:
            plotting_file = os.path.join(outdir,
                                         '{}-{}-centrimo-plot-data.txt'.format(
                                             transcription_factor, jaspar_id))
            # find DMC rank
            fe_df = read_and_sort_fisher_exact_centrimo_df(dmc_txt)
            dmc_rank = fe_df.index[fe_df["id"] == jaspar_id].tolist()[0]
            # find KOIN rank
            koin_rank = sorted_koin_df.index[
                sorted_koin_df["id"] == jaspar_id].tolist()[0]
            # find WT rank (from DMC sorted by E-value instead of
            # Fisher's E-value)
            wt_rank = sorted_dmc_df.index[
                sorted_dmc_df["id"] == jaspar_id].tolist()[0]
            # find number of peako motifs
            num_peako_motifs = len(rankings_ratio_sorted_df)
            # find number of peaks WT
            data_dmc_html = store_data_from_html(dmc_html)
            num_wt_peaks = data_dmc_html["sequence_db"]["count"]
            # find number of peaks KO
            num_ko_peaks = data_dmc_html["negative_sequence_db"]["count"]
            # find number of peaks KOIN
            data_koin_html = store_data_from_html(koin_html)
            num_koin_peaks = data_koin_html["sequence_db"]["count"]
            with open(plotting_file, 'w') as csv_file:
                writer = csv.writer(csv_file, delimiter=',',
                                    lineterminator='\n')
                # Experiment, DMC, KOIN, WT, peaKO, peaKO_wo_filtering,
                # motifs, WT_peaks, KO_peaks, KOIN_peaks
                writer.writerow(["Experiment", "DMC", "KOIN", "WT", "peaKO",
                                 "peaKO_wo_filtering",
                                 "num_motifs_peaKO_wo_filt",
                                 "num_motifs_DMC", "num_motifs_KOIN",
                                 "WT_peaks", "KO_peaks", "KOIN_peaks"])
                writer.writerow([transcription_factor, dmc_rank, koin_rank,
                                 wt_rank, peako_with_filtering_rank,
                                 peako_rank, num_peako_motifs,
                                 number_of_motifs, number_of_koin_motifs,
                                 num_wt_peaks, num_ko_peaks,
                                 num_koin_peaks])

    logging.info('Finished! :)')


if __name__ == '__main__':
    main()


# TO DO:
# change output plotting file names to match requirements of plotting scripts
# when printing plotting file, use tuples to group column names and values



#!/usr/bin/env python
from __future__ import print_function, division
from __future__ import absolute_import
import argparse
from bs4 import BeautifulSoup
import yaml
from collections import OrderedDict
import csv
import os
import re
import six
from six.moves import filter

parser = argparse.ArgumentParser(description='Add pooled MEME and DREME ' +
                                 'motif databases (Pipeline A, Pipeline B, ' +
                                 'and WT) to CentriMo calls. Must have ' +
                                 'already run pool-de-novo-motifs-before-' +
                                 'centrimo.py.')
parser.add_argument('-l', '--list', nargs='+',
                    help='list of CentriMo .html files', required=True)
parser.add_argument('-o', help='output directory for CentriMo', required=True)
args = parser.parse_args()


# PARAMETERS

centrimo_file_list = args.list
output_directory = args.o  # for CentriMo call output
# (new script will be created in $(pwd)/centrimo-neg_out)

# regexes to find dreme or meme motifs databases (.xml)
# and Pipeline A, Pipeline B, and WT files, and JASPAR db
regex = re.compile(r'dreme.xml|meme.xml')
koin_regex = re.compile(r'.*PipelineB.*')
dmc_regex = re.compile(r'.*PipelineA.*')
wt_regex = re.compile(r'.*WT.*')
db_regex = re.compile(r'.*JASPAR.*')


# FUNCTIONS

# Adapted from: http://stackoverflow.com/questions/13323976/
def parse_html(htmlfile):
    """
    Read an HTML file and return text between 'script' tags
    as string objects.
    """
    opened_file = open(htmlfile)
    html = opened_file.read()
    soup = BeautifulSoup(html, "html.parser")
    opened_file.close()
    script = soup.find('script')
    json_text = script.text.split('=', 1)[1].strip().rstrip(';')
    data = yaml.safe_load(json_text)  # loads into string objects
    return data


def change_to_pool_filename(filename):
    """Return corresponding full filepath of pooled motif file."""
    return (os.path.dirname(filename) + '/' +
            os.path.splitext(os.path.basename(filename))[0] +
            '-pool.xml')


# SCRIPT

cmd_dictionary = OrderedDict()
pooled_de_novo_motif_db_dictionary = OrderedDict()

for filename in centrimo_file_list:
    filebase = os.path.basename(filename)

    centrimo_data = parse_html(filename)
    # centrimo_data['cmd'][0] = "centrimo-neg"  # change from 'centrimo'
    centrimo_data['cmd'][6] = (output_directory + '/centrimo-' +
                               str(centrimo_file_list.index(filename)) + '/')

    # check filepath for regexes defined above
    # if present, then append meme-pool.xml and dreme-pool.xml to a
    # list or dictionary
    # XXX should simplify this section; lots of repeated code
    if dmc_regex.search(filename):
        cmd = [change_to_pool_filename(i) if regex.search(i)
               else i for i in centrimo_data['cmd']]
        pooled_de_novo_motif_db_dictionary['DMC'] = [
            change_to_pool_filename(i) for i in centrimo_data['cmd']
            if regex.search(i)]
        simple_file = 'DMC'
    elif koin_regex.search(filename):
        cmd = [change_to_pool_filename(i) if regex.search(i)
               else i for i in centrimo_data['cmd']]
        pooled_de_novo_motif_db_dictionary['KOIN'] = [
            change_to_pool_filename(i) for i in centrimo_data['cmd']
            if regex.search(i)]
        simple_file = 'KOIN'
    elif wt_regex.search(filename):
        cmd = [change_to_pool_filename(i) if regex.search(i)
               else i for i in centrimo_data['cmd']]
        pooled_de_novo_motif_db_dictionary['WT'] = [
            change_to_pool_filename(i) for i in centrimo_data['cmd']
            if regex.search(i)]
        simple_file = 'WT'
    else:
        cmd = centrimo_data['cmd']
        simple_file = (os.path.splitext(filebase)[0] +
                       str(centrimo_file_list.index(filename)))
        # remove extension

    cmd_dictionary[simple_file] = cmd


# insert motif databases (.xml files) from DMC, KOIN, WT runs
for key, cmd in six.iteritems(cmd_dictionary):
    if key == 'DMC':
        insert_motif_dbs = (pooled_de_novo_motif_db_dictionary['KOIN'] +
                            pooled_de_novo_motif_db_dictionary['WT'])
        for i in insert_motif_dbs:
            cmd.insert(cmd.index(list(filter(db_regex.match, cmd))[0]), i)
    elif key == 'KOIN':
        insert_motif_dbs = (pooled_de_novo_motif_db_dictionary['DMC'] +
                            pooled_de_novo_motif_db_dictionary['WT'])
        for i in insert_motif_dbs:
            cmd.insert(cmd.index(list(filter(db_regex.match, cmd))[0]), i)
    elif key == 'WT':
        insert_motif_dbs = (pooled_de_novo_motif_db_dictionary['DMC'] +
                            pooled_de_novo_motif_db_dictionary['KOIN'])
        for i in insert_motif_dbs:
            cmd.insert(cmd.index(list(filter(db_regex.match, cmd))[0]), i)
    else:
        insert_motif_dbs = (pooled_de_novo_motif_db_dictionary['KOIN'] +
                            pooled_de_novo_motif_db_dictionary['DMC'] +
                            pooled_de_novo_motif_db_dictionary['WT'])
        for i in insert_motif_dbs:
            cmd.insert(cmd.index(list(filter(db_regex.match, cmd))[0]), i)
    cmd_dictionary[key] = ' '.join(cmd)


# print commands to file
output_file = os.path.join('.', 'commands-pooled-de-novo.sh')

with open(output_file, 'w') as csv_file:
    writer = csv.writer(csv_file, delimiter='\t', lineterminator='\n\n')
    writer.writerow(["#!/usr/bin/env bash"])
    for _, cmd in cmd_dictionary.items():
        writer.writerow([cmd])



#!/usr/bin/env python

import argparse
import os
import logging
import sys
import subprocess
import json

from pkg_resources import resource_filename
#from .version import __version__

snakefile = resource_filename(__name__, "data/snakefile")


def get_basename_without_exts(file_path):
    # adapted from: https://stackoverflow.com/questions/678236/
    file_basename = os.path.basename(file_path)
    filename_without_extension = file_basename.split('.')[0]
    return filename_without_extension


def get_sample_name(sample):
    if sample.endswith(('.bam.gz', '.bam')):
        if sample.endswith('.bam'):
            sample_name = get_basename_without_exts(sample)
    else:
        sys.exit("{} must end in .bam or .bam.gz".format(sample))
    return sample_name


def check_and_create_symlink(directory, original_filepath, symlink_filepath):
    if not os.path.isdir(directory):
        subprocess.run("mkdir {}".format(directory), shell=True)
    if not os.path.isfile(symlink_filepath):
        subprocess.run("ln -s {} {}".format(original_filepath,
                                            symlink_filepath),
                       shell=True)
    else:
        logging.info("{} already exists".format(symlink_filepath))


def main():
    logging.basicConfig(format='%(asctime)s >peaKO-workflow: %(message)s',
                        datefmt='%m/%d/%Y %I:%M:%S %p', level=logging.INFO)

    parser = argparse.ArgumentParser(
        prog='peako',
        description="""
PeaKO ranks motifs from ChIP-seq experiments performed with knockout
controls. This peaKO workflow processes a pair of wild-type/knockout
BAM files and returns motif rankings. PeaKO computes its ranking
metric by comparing CentriMo results across two different ways of
differentially accounting for the knockout dataset (Pipeline A vs.
Pipeline B; see manuscript for details).
        """,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""

Example CentriMo HTML and TXT files are available at: \
https://doi.org/10.5281/zenodo.3338330.

Usage example:
$ peako ~/workdir/peako-workflow/ \

        ~/analysis/tead4-wt.bam   \

        ~/analysis/tead4-ko.bam   \

        mouse                     \

        ~/mm10/chrsize.txt         \

        ~/trf/mm10-masked.fa       \

        ~/motif_database/JASPAR_CORE_2016_vertebrates.meme

If you found peaKO useful, please cite:

Denisko D, Viner C, Hoffman MM. Motif elucidation in ChIP-seq datasets with a
knockout control. BioRxiv <ID> [Preprint]. 2019. \
Available from: https://doi.org/<ID>
        """)

    parser.add_argument('workdir', metavar="outdir",
                        help='output directory')
    parser.add_argument('wt_bam', metavar="wt-bam",
                        help='wild-type sample BAM file')
    parser.add_argument('ko_bam', metavar="ko-bam",
                        help='knockout sample BAM file')

    parser.add_argument('organism', metavar="organism",
                        help='name of organism (mouse or human)')
    parser.add_argument('chr_sizes', metavar="chr-sizes",
                        help='chromosome sizes file of reference genome (TXT)')
    parser.add_argument('trf_genome', metavar="trf-masked-genome",
                        help='TRF masked reference genome file (FASTA)')
    parser.add_argument('motif_database', metavar="motif-database",
                        help='motif database (MEME)')

    # general
#    parser.add_argument('-V', '--version', action='version',
#                        version=__version__)

    # peaKO submodule
    parser.add_argument('-j', dest='jaspar_id',
                        help='transcription factor motif JASPAR ' +
                        'identifier (e.g. MA0083.3)')
    parser.add_argument('-m', dest='motif',
                        help='transcription factor motif common name ' +
                        '(e.g. SRF)')
    parser.add_argument('--extra', action="store_true",
                        help='output all intermediate peaKO files for ' +
                        'plotting')
    parser.add_argument('--pickle', action="store_true",
                        help='use pickled peaKO dictionaries from previous ' +
                        'run')

    # snakemake
    parser.add_argument('--sm-build-envs', action="store_true",
                        help='build conda environments for workflow and ' +
                        'exit (requires internet connection)')
    parser.add_argument('--sm-cluster-config', dest='sm_cluster_config',
                        help='snakemake cluster configuration file (JSON)')
    # parser.add_argument('--sm-config', dest='sm_config',
    #                     help='snakemake configuration file (JSON)')

    args = parser.parse_args()
    # logging.info('Command line arguments: {}'.format(args))

    workdir = args.workdir
    wt_bam = args.wt_bam
    ko_bam = args.ko_bam
    organism = args.organism
    motif_database = args.motif_database
    chr_sizes = args.chr_sizes
    trf_genome = args.trf_genome
    peako_jaspar_id = args.jaspar_id
    peako_motif = args.motif
    peako_extra = args.extra
    peako_pickle = args.pickle

    script_path = os.path.dirname(os.path.realpath(__file__))

    # check that directory and all files exist
    if not os.path.isdir(workdir):
        sys.exit("{} is not a directory... exiting!".format(workdir))

    file_list = [wt_bam, ko_bam, chr_sizes, trf_genome]
    for f in file_list:
        if not os.path.isfile(f):
            sys.exit("{} is not a file... exiting!".format(f))

    # derive sample names
    wt_sample = get_sample_name(wt_bam)
    ko_sample = get_sample_name(ko_bam)
    pipb_sample = "-".join([wt_sample, ko_sample])

    # create symlinks in workdir/data and workdir/reference_files
    # (if not already present)
    data_dir = os.path.join(workdir, "data")
    ref_dir = os.path.join(workdir, "reference_files")

    wt_symlink = os.path.join(data_dir, "{}-wt.bam".format(wt_sample))
    ko_symlink = os.path.join(data_dir, "{}-ko.bam".format(ko_sample))
    motif_db_symlink = os.path.join(ref_dir, os.path.basename(motif_database))
    chr_sizes_symlink = os.path.join(ref_dir, os.path.basename(chr_sizes))
    trf_genome_symlink = os.path.join(ref_dir, os.path.basename(trf_genome))

    check_and_create_symlink(data_dir, wt_bam, wt_symlink)
    check_and_create_symlink(data_dir, ko_bam, ko_symlink)
    check_and_create_symlink(ref_dir, motif_database, motif_db_symlink)
    check_and_create_symlink(ref_dir, chr_sizes, chr_sizes_symlink)
    check_and_create_symlink(ref_dir, trf_genome, trf_genome_symlink)

    # associate organism with mm or hs for MACS2
    if organism == "mouse":
        macs2_genome_size = "mm"
    elif organism == "human":
        macs2_genome_size = "hs"
    else:
        sys.exit("{} is not a valid organism for ".format(organism) +
                 "this workflow (must choose mouse or human)... exiting!")

    # convert peako args to a string
    peako_params = []
    if peako_extra:
        peako_params.append("--extra")
    if peako_pickle:
        peako_params.append("--pickle")
    if peako_jaspar_id:
        peako_params.append("-j {}".format(peako_jaspar_id))
    if peako_motif:
        peako_params.append("-m {}".format(peako_motif))

    if peako_params:
        peako_params_command = " ".join(peako_params)
    else:
        peako_params_command = ""
    
    # convert snakemake args to string
    # sm_additional_params = []
    # if sm_unlock:
    #     sm_additional_params.append("--unlock")

    # write everything to JSON file
    config_file = os.path.join(workdir, "config.json")
    with open(config_file, "w") as f:
        json.dump({'sample_wt_prefix': wt_sample,
                   'sample_ko_prefix': ko_sample,
                   'sample_pipb_prefix': pipb_sample,
                   'macs2_genome_size': macs2_genome_size,
                   'macs2_q_value': 0.05,
                   'chr_sizes': ("reference_files/{}"
                                 .format(os.path.basename(chr_sizes))),
                   'trf_masked_genome': ("reference_files/{}"
                                         .format(os.path
                                                 .basename(trf_genome))),
                   'motif_database': ("reference_files/{}"
                                      .format(os.path
                                              .basename(motif_database))),
                   'min_meme_motif_width': 7,
                   'max_meme_motif_width': 12,
                   'peako_params': peako_params_command,
                   'script_path': script_path},
                  f, indent=4)

    # build conda environments and exit
    if args.sm_build_envs:
        subprocess.run('snakemake -s {} -d {} '.format(snakefile, workdir) +
                       '--use-conda --create-envs-only',
                       shell=True)
        sys.exit("Built Conda environments for peaKO. You can now run " +
                 "it on a cluster without internet if you wish.")

    # source activate conda enrironment (has to already be built)
    if args.sm_cluster_config:
        cluster_command = ('"sbatch -p {cluster.partition} ' +
                           '-t {cluster.time} --mem {cluster.mem} ' +
                           '-J {cluster.name} -o {cluster.log}"')
        subprocess.run('snakemake -s {} -d {} '.format(snakefile, workdir) +
                       '--configfile {} '.format(config_file) +
                       '-j 4 --latency-wait 60 --use-conda ' +
                       '--cluster-config {} '.format(args.sm_cluster_config) +
                       '--cluster {}'.format(cluster_command),
                       shell=True)
    else:
        subprocess.run('snakemake -s {} -d {} '.format(snakefile, workdir) +
                       '--configfile {} --use-conda '.format(config_file) +
                       '--latency-wait 60',
                       shell=True)


if __name__ == '__main__':
    main()

# TO DO:
# change verbosity of all programs at once
# maybe give option to provide custom config and skip over all of these steps



#!/usr/bin/env python
from __future__ import print_function, division
from __future__ import absolute_import
import argparse
import os
from xml.etree import ElementTree as et
import logging
logging.basicConfig(format='%(asctime)s %(message)s',
                    datefmt='%m/%d/%Y %I:%M:%S %p', level=logging.INFO)

parser = argparse.ArgumentParser(description='Modify ID of MEME and DREME ' +
                                 'de novo motifs to preserve uniqueness ' +
                                 'upon pooling DMC and KOIN pipelines.')
parser.add_argument('-d', help='full path of DMC directory')
parser.add_argument('-k', help='full path of KOIN directory')
parser.add_argument('-w', help='full path of WT directory')

args = parser.parse_args()

# dreme and meme files (and folders) can just be found based off
# of this because they always follow the same pattern
# (when output by centrimo)

meme_xml_dmc = args.d + '/meme_out/meme.xml'
dreme_xml_dmc = args.d + '/dreme_out/dreme.xml'

meme_xml_koin = args.k + '/meme_out/meme.xml'
dreme_xml_koin = args.k + '/dreme_out/dreme.xml'

meme_xml_wt = args.w + '/meme_out/meme.xml'
dreme_xml_wt = args.w + '/dreme_out/dreme.xml'

# PARAMETERS

output_directory_meme_dmc = os.path.dirname(meme_xml_dmc)
output_directory_meme_koin = os.path.dirname(meme_xml_koin)
output_directory_meme_wt = os.path.dirname(meme_xml_wt)

output_directory_dreme_dmc = os.path.dirname(dreme_xml_dmc)
output_directory_dreme_koin = os.path.dirname(dreme_xml_koin)
output_directory_dreme_wt = os.path.dirname(dreme_xml_wt)


# FUNCTIONS


def replace_meme_id(meme_root, suffix):
    for motif in meme_root.iter('motif'):
        new_id = motif.attrib['name'] + suffix
        motif.set('name', new_id)


def add_motifs_to_dct(dreme_root, label):
    for motif in dreme_root.iter('motif'):
        motif_id = motif.attrib['seq']
        if motif_id in dreme_dct:
            dreme_dct[motif_id].append(label)
        else:
            dreme_dct[motif_id] = [label]


def append_suffix_dreme_id(dreme_root):
    for motif in dreme_root[1].findall('motif'):
        # XXX assume that 'motifs' is second layer
        motif_id = motif.attrib['seq']
        if dreme_dct[motif_id][-1] == 'used':
            dreme_root[1].remove(motif)
        else:
            suffix = '-' + '-'.join(dreme_dct[motif_id])
            new_id = motif_id + suffix
            motif.set('seq', new_id)
            dreme_dct[motif_id].append('used')


# BUILD TREE
def get_root(xml):
    tree = et.parse(xml)
    return tree.getroot()


tree_meme_dmc = et.parse(meme_xml_dmc)
root_meme_dmc = tree_meme_dmc.getroot()

tree_meme_koin = et.parse(meme_xml_koin)
root_meme_koin = tree_meme_koin.getroot()

tree_meme_wt = et.parse(meme_xml_wt)
root_meme_wt = tree_meme_wt.getroot()

tree_dreme_dmc = et.parse(dreme_xml_dmc)
root_dreme_dmc = tree_dreme_dmc.getroot()

tree_dreme_koin = et.parse(dreme_xml_koin)
root_dreme_koin = tree_dreme_koin.getroot()

tree_dreme_wt = et.parse(dreme_xml_wt)
root_dreme_wt = tree_dreme_wt.getroot()

logging.info('Built trees and got roots!')

# Alter MEME motif labels
replace_meme_id(root_meme_dmc, '-DMC')
replace_meme_id(root_meme_koin, '-KOIN')
replace_meme_id(root_meme_wt, '-WT')
logging.info('Renamed MEME motifs.')

# Create dictionary of DREME motif labels
dreme_dct = {}
add_motifs_to_dct(root_dreme_dmc, 'DMC')
add_motifs_to_dct(root_dreme_koin, 'KOIN')
add_motifs_to_dct(root_dreme_wt, 'WT')

# Alter DREME motif labels
append_suffix_dreme_id(root_dreme_dmc)
append_suffix_dreme_id(root_dreme_koin)
append_suffix_dreme_id(root_dreme_wt)
logging.info('Renamed DREME motifs and removed duplicates.')

# Comment:
# XXX It would be nice to output everything to the same file, but then
# I would have to insert between motif tags. This is doable BUT I think
# it's too confusing because of the information under 'model', unless
# we remove 'model' and 'run_time' tags completely.

# could change 'last_mod_dat' attribute of first child (tag 'model') of root
tree_meme_dmc.write(output_directory_meme_dmc + '/meme-pool.xml')
tree_meme_koin.write(output_directory_meme_koin + '/meme-pool.xml')
tree_meme_wt.write(output_directory_meme_wt + '/meme-pool.xml')

tree_dreme_dmc.write(output_directory_dreme_dmc + '/dreme-pool.xml')
tree_dreme_koin.write(output_directory_dreme_koin + '/dreme-pool.xml')
tree_dreme_wt.write(output_directory_dreme_wt + '/dreme-pool.xml')
logging.info('Wrote new trees to files. Check output directories!')



#!/usr/bin/env python

from setuptools import setup
from peako import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name="peako",
      version=__version__,
      description="A package for discovering motifs in ChIP-seq datasets with knockout controls",
      author="Danielle Denisko",
      author_email="danielle.denisko@mail.utoronto.ca",
      license='GPLv3',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://bitbucket.org/hoffmanlab/peako",
      packages=['peako'],
      package_data={'peako': ['data/*']},
      entry_points = {
        'console_scripts': [
            'peako=peako.peako_snakemake_wrapper:main',
            ],
        },
      zip_safe=False)
